#!/usr/bin/env python3
#
# Copyright (c) 2016-2026, Babak Farrokhi
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import errno
import getopt
import os
import socket
import ssl
import struct
import sys
import time
from statistics import stdev
from typing import Any

import dns.edns
import dns.exception
import dns.flags
import dns.message
import dns.query
import dns.rcode
import dns.rdata
import dns.rdataclass
import dns.rdatatype
import dns.resolver
import httpx

from dnsdiag.dns import PROTO_UDP, PROTO_TCP, PROTO_TLS, PROTO_HTTPS, PROTO_QUIC, PROTO_HTTP3, proto_to_text, \
    get_default_port, valid_rdatatype, CustomSocket
import dnsdiag.shared as shared
from dnsdiag.shared import __version__, valid_hostname, unsupported_feature, random_string, die, err, \
    set_protocol_exclusive, parse_ip_address, setup_signal_handler, resolve_server_address

__author__ = 'Babak Farrokhi (babak@farrokhi.net)'
__license__ = 'BSD'
__progname__ = os.path.basename(sys.argv[0])


def usage(exit_code: int = 0) -> None:
    print("""%s version %s
Usage: %s [-346aDeEFhLmqnrvTQxXH] [-i interval] [-w wait] [-p dst_port] [-P src_port] [-S src_ip]
       %s [-c count] [-t qtype] [-C class] [-s server] [--ecs client_subnet] hostname

  -h, --help        Show this help message
  -q, --quiet       Suppress output
  -v, --verbose     Print the full DNS response
  -s, --server      Specify the DNS server to use (default: first entry from /etc/resolv.conf)
  -p, --port        Specify the DNS server port number (default: 53 for TCP/UDP, 853 for TLS)
  -T, --tcp         Use TCP as the transport protocol
  -X, --tls         Use TLS as the transport protocol
  -H, --doh         Use HTTPS as the transport protocol (DoH)
  -3, --http3       Use HTTP/3 as the transport protocol (DoH3)
  -Q, --quic        Use QUIC as the transport protocol (DoQ)
  -4, --ipv4        Use IPv4 as the network protocol
  -6, --ipv6        Use IPv6 as the network protocol
  -P, --srcport     Specify the source port number for the query (default: 0)
  -S, --srcip       Specify the source IP address for the query (default: default interface address)
  -c, --count       Number of requests to send (default: 10, 0 for unlimited)
  -r, --norecurse   Enforce a non-recursive query by clearing the RD (recursion desired) bit
  -m, --cache-miss  Force cache miss measurement by prepending a random hostname
  -w, --wait        Maximum wait time for a reply (default: 2 seconds)
  -i, --interval    Time interval between requests (default: 1 second)
  -t, --type        DNS request record type (default: A)
  -L, --ttl         Display the response TTL (if present)
  -C, --class       DNS request record class (default: IN)
  -a, --answer      Display the first matching answer in rdata, if applicable
  -e, --edns        Enable EDNS0 and set its options
  -E, --ede         (Ignored - EDE messages now always displayed when present)
  -n, --nsid        Enable the NSID bit to retrieve resolver identification (implies EDNS)
      --cookie      Display EDNS cookies when present (implies EDNS)
  -D, --dnssec      Enable the DNSSEC desired flag (implies EDNS)
      --ecs         Set EDNS Client Subnet option (format: IP/prefix, e.g., 192.168.1.0/24) (implies EDNS)
  -F, --flags       Display response flags
  -x, --expert      Display additional information (implies --ttl, --flags)
""" % (__progname__, __version__, __progname__, ' ' * len(__progname__)))
    sys.exit(exit_code)


def main() -> None:
    setup_signal_handler()

    if len(sys.argv) == 1:
        usage()

    dns.rdata.load_all_types()  # type: ignore[no-untyped-call]
    # defaults
    rdatatype = 'A'
    rdata_class = dns.rdataclass.from_text('IN')
    count = 10
    timeout = 2
    interval = 1.0
    quiet = False
    verbose = False
    show_flags = False
    show_cookie = False
    dnsserver = None  # do not try to use system resolver by default
    proto = PROTO_UDP
    dst_port = get_default_port(proto)
    use_default_dst_port = True
    src_port = 0
    src_ip = None
    use_edns = False
    want_nsid = False
    want_dnssec = False
    client_subnet = None
    show_ttl = False
    force_miss = False
    show_answer = False
    request_flags = dns.flags.from_text('RD')
    af = None
    af_ipv4_set = False
    af_ipv6_set = False
    proto_option_set: str | None = None
    qname = 'wikipedia.org'

    try:
        opts, args = getopt.getopt(sys.argv[1:], "qhc:s:t:w:i:vp:P:S:TQ346meDFXHrnEC:Lxa",
                                   ["help", "count=", "server=", "quiet", "type=", "wait=", "interval=", "verbose",
                                    "port=", "srcip=", "tcp", "ipv4", "ipv6", "cache-miss", "srcport=", "edns",
                                    "dnssec", "flags", "norecurse", "tls", "doh", "nsid", "ede", "class=", "ttl",
                                    "expert", "answer", "quic", "http3", "ecs=", "cookie"])
    except getopt.GetoptError as getopt_err:
        err(str(getopt_err))
        usage(1)

    for o, a in opts:
        if o in ("-h", "--help"):
            usage()

    if args and len(args) == 1:
        qname = args[0]
        if not valid_hostname(qname, allow_underscore=True):
            die(f"ERROR: invalid hostname: {qname}")
    else:
        usage(1)

    for o, a in opts:
        if o in ("-c", "--count"):
            if a.isdigit():
                count = abs(int(a))
            else:
                die(f"ERROR: invalid count of requests: {a}")

        elif o in ("-v", "--verbose"):
            verbose = True

        elif o in ("-s", "--server"):
            dnsserver = a

        elif o in ("-q", "--quiet"):
            quiet = True
            verbose = False

        elif o in ("-w", "--wait"):
            try:
                timeout = int(a)
                if timeout < 0:
                    die(f"ERROR: wait time must be non-negative: {a}")
            except ValueError:
                die(f"ERROR: invalid wait time value: {a}")

        elif o in ("-a", "--answer"):
            show_answer = True

        elif o in ("-x", "--expert"):
            show_flags = True
            show_ttl = True

        elif o in ("-m", "--cache-miss"):
            force_miss = True

        elif o in ("-i", "--interval"):
            try:
                interval = float(a)
                if interval < 0:
                    die(f"ERROR: interval must be non-negative: {a}")
            except ValueError:
                die(f"ERROR: invalid interval value: {a}")

        elif o in ("-L", "--ttl"):
            show_ttl = True

        elif o in ("-t", "--type"):
            rdatatype = a

        elif o in ("-C", "--class"):
            try:
                rdata_class = dns.rdataclass.from_text(a)
            except dns.rdataclass.UnknownRdataclass:
                die(f"ERROR: invalid RR class: {a}")

        elif o in ("-T", "--tcp"):
            proto, proto_option_set = set_protocol_exclusive(PROTO_TCP, o, proto_option_set)
            if use_default_dst_port:
                dst_port = get_default_port(proto)

        elif o in ("-X", "--tls"):
            proto, proto_option_set = set_protocol_exclusive(PROTO_TLS, o, proto_option_set)
            if use_default_dst_port:
                dst_port = get_default_port(proto)

        elif o in ("-H", "--doh"):
            proto, proto_option_set = set_protocol_exclusive(PROTO_HTTPS, o, proto_option_set)
            if use_default_dst_port:
                dst_port = get_default_port(proto)

        elif o in ("-Q", "--quic"):
            proto, proto_option_set = set_protocol_exclusive(PROTO_QUIC, o, proto_option_set)
            if use_default_dst_port:
                dst_port = get_default_port(proto)

        elif o in ("-3", "--http3"):
            proto, proto_option_set = set_protocol_exclusive(PROTO_HTTP3, o, proto_option_set)
            if use_default_dst_port:
                dst_port = get_default_port(proto)

        elif o in ("-4", "--ipv4"):
            if af_ipv6_set:
                die("ERROR: cannot specify both -4 and -6")
            af = socket.AF_INET
            af_ipv4_set = True

        elif o in ("-6", "--ipv6"):
            if af_ipv4_set:
                die("ERROR: cannot specify both -4 and -6")
            af = socket.AF_INET6
            af_ipv6_set = True

        elif o in ("-e", "--edns"):
            use_edns = True

        elif o in ("-n", "--nsid"):
            use_edns = True  # required
            want_nsid = True

        elif o in ("-r", "--norecurse"):
            request_flags = dns.flags.from_text('')

        elif o in ("-D", "--dnssec"):
            use_edns = True  # required
            want_dnssec = True

        elif o in ("-F", "--flags"):
            show_flags = True

        elif o in ("-E", "--ede"):
            # Kept for compatibility - EDE now always displayed when present
            pass

        elif o == "--cookie":
            use_edns = True  # required
            show_cookie = True

        elif o in ("-p", "--port"):
            try:
                dst_port = int(a)
                if not (0 < dst_port <= 65535):
                    die(f"ERROR: port must be between 1 and 65535: {a}")
                use_default_dst_port = False
            except ValueError:
                die(f"ERROR: invalid port value: {a}")

        elif o in ("-P", "--srcport"):
            try:
                src_port = int(a)
                if not (0 <= src_port <= 65535):
                    die(f"ERROR: source port must be between 0 and 65535: {a}")
                if src_port < 1024 and not quiet:
                    err("WARNING: Source ports below 1024 are only available to superuser")
            except ValueError:
                die(f"ERROR: invalid source port value: {a}")

        elif o in ("-S", "--srcip"):
            src_ip = parse_ip_address(a)

        elif o == "--ecs":
            client_subnet = a
            use_edns = True  # ECS requires EDNS

        else:
            usage(1)

    if src_ip is not None:
        if af is not None:
            parse_ip_address(src_ip, family=af)
        else:
            af = socket.AF_INET6 if ':' in src_ip else socket.AF_INET

    if src_port > 0:
        # For UDP with a fixed source port, SO_REUSEADDR lets the socket bind the same
        # port again between queries. TCP avoids this entirely via connection reuse (tcp_sock).
        dns.query.socket_factory = CustomSocket

    # Use system DNS server if parameter is not specified
    # remember not all systems have /etc/resolv.conf (i.e. Android)
    if dnsserver is None:
        dnsserver = str(dns.resolver.get_default_resolver().nameservers[0])

    dnsserver_hostname = dnsserver  # keep original name for display and SNI
    dnsserver_ip = resolve_server_address(dnsserver, af)

    response_time = []
    i = 0

    # validate RR type
    if not valid_rdatatype(rdatatype):
        die(f'ERROR: invalid record type: {rdatatype}')

    # For TCP with a fixed source port, keep one connection alive across all queries to avoid
    # TIME_WAIT exhaustion — the OS cannot reuse an identical 4-tuple while the old one is in
    # TIME_WAIT, even with SO_REUSEADDR.  This is also the RFC 7766-recommended pattern.
    tcp_sock = None
    if proto is PROTO_TCP and src_port > 0:
        af_tcp = socket.AF_INET6 if ':' in dnsserver_ip else socket.AF_INET
        tcp_sock = socket.socket(af_tcp, socket.SOCK_STREAM)
        tcp_sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        source = (src_ip or ('::' if af_tcp == socket.AF_INET6 else '0.0.0.0'), src_port)
        tcp_sock.bind(source)
        tcp_sock.connect((dnsserver_ip, dst_port))
        tcp_sock.setblocking(False)

    # Display the hostname if it differs from the resolved IP, otherwise just the IP
    server_display = dnsserver_hostname if dnsserver_hostname != dnsserver_ip else dnsserver_ip
    # Wrap IPv6 addresses in brackets for better readability
    if ':' in server_display and not server_display.startswith('['):
        server_display = f"[{server_display}]"
    print("%s DNS: %s:%d, hostname: %s, proto: %s, class: %s, type: %s, flags: [%s]" %
          (__progname__, server_display, dst_port, qname, proto_to_text(proto), dns.rdataclass.to_text(rdata_class),
           rdatatype, dns.flags.to_text(request_flags)), flush=True)

    while not shared.shutdown:

        if 0 < count <= i:
            break
        else:
            i += 1

        if force_miss:
            fqdn = "_dnsdiag_%s_.%s" % (random_string(8, 8), qname)
        else:
            fqdn = qname

        if use_edns:
            edns_options: list[Any] = []
            if want_nsid:
                edns_options.append(dns.edns.GenericOption(dns.edns.NSID, b''))
            if client_subnet:
                try:
                    ecs_option = dns.edns.ECSOption.from_text(client_subnet)
                    edns_options.append(ecs_option)
                except Exception as e:
                    die(f"ERROR: invalid ECS format '{client_subnet}': {e}")
            if show_cookie:
                # Send a client cookie (8 random bytes as per RFC 7873)
                client_cookie = os.urandom(8)
                edns_options.append(dns.edns.CookieOption(client_cookie, b''))

            query = dns.message.make_query(fqdn, rdatatype, rdata_class, flags=request_flags,
                                           use_edns=True, want_dnssec=want_dnssec, payload=1232,
                                           options=edns_options)
        else:
            query = dns.message.make_query(fqdn, rdatatype, rdata_class, flags=request_flags,
                                           use_edns=False, want_dnssec=False)

        try:
            stime = time.perf_counter()
            if proto is PROTO_UDP:
                answers = dns.query.udp(query, dnsserver_ip, timeout=timeout, port=dst_port,
                                        source=src_ip, source_port=src_port, ignore_unexpected=True)
            elif proto is PROTO_TCP:
                if tcp_sock:
                    # TODO: reconnect on connection drop (currently counted as packet loss)
                    answers = dns.query.tcp(query, dnsserver_ip, timeout=timeout, port=dst_port,
                                            source=src_ip, source_port=src_port, sock=tcp_sock)
                else:
                    answers = dns.query.tcp(query, dnsserver_ip, timeout=timeout, port=dst_port,
                                            source=src_ip, source_port=src_port)
            elif proto is PROTO_TLS:
                if hasattr(dns.query, 'tls'):
                    try:
                        # Use resolved IP for connection, but provide hostname for SNI/certificate validation
                        server_hostname = dnsserver_hostname if dnsserver_hostname != dnsserver_ip else None
                        answers = dns.query.tls(query, dnsserver_ip, timeout=timeout, port=dst_port,
                                                source=src_ip, source_port=src_port,
                                                server_hostname=server_hostname)
                    except dns.exception.Timeout:
                        if not quiet:
                            print("Request timeout", flush=True)
                        continue
                    except (ConnectionResetError, BrokenPipeError):
                        if not quiet:
                            print("Connection closed unexpectedly", flush=True)
                        continue
                    except socket.gaierror:
                        if not quiet:
                            print("Name resolution failed", flush=True)
                        continue
                    except ssl.SSLCertVerificationError as e:
                        die(f"Certificate verification failed: {e}")
                    except ssl.SSLError:
                        if not quiet:
                            print("Connection failed", flush=True)
                        continue
                else:
                    unsupported_feature("DNS-over-TLS")

            elif proto is PROTO_HTTPS:
                if hasattr(dns.query, 'https'):
                    try:
                        # For hostnames, construct full URL; for IPs, use IP directly
                        if dnsserver_hostname != dnsserver_ip:
                            https_server = f"https://{dnsserver_hostname}/dns-query"
                        else:
                            https_server = dnsserver_ip
                        answers = dns.query.https(query, https_server, timeout=timeout, port=dst_port,
                                                  source=src_ip, source_port=src_port,
                                                  http_version=dns.query.HTTPVersion.HTTP_2)
                    except dns.query.NoDOH:
                        die("ERROR: python httpx module not available")
                    except httpx.ConnectError:
                        if not quiet:
                            print("Connection refused", flush=True)
                        continue
                    except (ConnectionResetError, BrokenPipeError):
                        if not quiet:
                            print("Connection closed unexpectedly", flush=True)
                        continue
                    except socket.gaierror:
                        if not quiet:
                            print("Name resolution failed", flush=True)
                        continue
                    except ssl.SSLCertVerificationError as e:
                        die(f"Certificate verification failed: {e}")
                    except (ssl.SSLError, httpx.WriteTimeout, httpx.StreamError, httpx.ProtocolError):
                        if not quiet:
                            print("Connection failed", flush=True)
                        continue
                else:
                    unsupported_feature("DNS-over-HTTPS (DoH)")

            elif proto is PROTO_HTTP3:
                if hasattr(dns.query, 'https') and hasattr(dns.query, 'HTTPVersion') and hasattr(dns.query.HTTPVersion, 'H3'):
                    try:
                        # For hostnames, construct full URL; for IPs, use IP directly
                        if dnsserver_hostname != dnsserver_ip:
                            https_server = f"https://{dnsserver_hostname}/dns-query"
                        else:
                            https_server = dnsserver_ip
                        answers = dns.query.https(query, https_server, timeout=timeout, port=dst_port,
                                                  source=src_ip, source_port=src_port,
                                                  http_version=dns.query.HTTPVersion.H3)
                    except ConnectionRefusedError:
                        if not quiet:
                            print("Connection refused", flush=True)
                        continue
                    except (ConnectionResetError, BrokenPipeError):
                        if not quiet:
                            print("Connection closed unexpectedly", flush=True)
                        continue
                    except socket.gaierror:
                        if not quiet:
                            print("Name resolution failed", flush=True)
                        continue
                    except ssl.SSLCertVerificationError as e:
                        die(f"Certificate verification failed: {e}")
                    except (ssl.SSLError, httpx.WriteTimeout, httpx.StreamError, httpx.ProtocolError):
                        if not quiet:
                            print("Connection failed", flush=True)
                        continue
                    except Exception as e:
                        # Catch QUIC-specific exceptions (UnexpectedEOF, etc.)
                        if e.__class__.__name__ in ('UnexpectedEOF', 'QuicConnectionError', 'H3Error'):
                            if not quiet:
                                print("Connection closed unexpectedly", flush=True)
                            continue
                        else:
                            raise
                else:
                    unsupported_feature("DNS-over-HTTPS/3 (DoH3)")

            elif proto is PROTO_QUIC:
                if hasattr(dns.query, 'quic'):
                    try:
                        # Use resolved IP for connection, but provide hostname for SNI/certificate validation
                        server_hostname = dnsserver_hostname if dnsserver_hostname != dnsserver_ip else None
                        answers = dns.query.quic(query, dnsserver_ip, timeout=timeout, port=dst_port,
                                                 source=src_ip, source_port=src_port,
                                                 server_hostname=server_hostname)
                    except dns.exception.Timeout:
                        if not quiet:
                            print("Request timeout", flush=True)
                        continue
                    except ConnectionRefusedError:
                        if not quiet:
                            print("Connection refused", flush=True)
                        continue
                    except (ConnectionResetError, BrokenPipeError):
                        if not quiet:
                            print("Connection closed unexpectedly", flush=True)
                        continue
                    except socket.gaierror:
                        if not quiet:
                            print("Name resolution failed", flush=True)
                        continue
                    except ssl.SSLCertVerificationError as e:
                        die(f"Certificate verification failed: {e}")
                    except ssl.SSLError:
                        if not quiet:
                            print("Connection failed", flush=True)
                        continue
                    except Exception as e:
                        # Catch QUIC-specific exceptions
                        if e.__class__.__name__ in ('UnexpectedEOF', 'QuicConnectionError', 'StreamFinishedError'):
                            if not quiet:
                                print("Connection closed unexpectedly", flush=True)
                            continue
                        else:
                            raise
                else:
                    unsupported_feature("DNS-over-QUIC (DoQ)")

            etime = time.perf_counter()
        except dns.resolver.NoNameservers as e:
            if not quiet:
                err("No response to DNS request")
                if verbose:
                    err(f"ERROR: {e}")
            sys.exit(1)
        except (httpx.ConnectTimeout, dns.exception.Timeout):
            if not quiet:
                print("Request timeout", flush=True)
        except httpx.ReadTimeout:
            if not quiet:
                print("Read timeout", flush=True)
        except EOFError:
            if not quiet:
                err("Connection closed by server")
        except PermissionError:
            if not quiet:
                die("ERROR: permission denied")
            else:
                sys.exit(1)
        except OSError as e:
            if e.errno == errno.EHOSTUNREACH:
                if not quiet:
                    print("No route to host", flush=True)
            elif e.errno == errno.ENETUNREACH:
                if not quiet:
                    print("Network unreachable", flush=True)
            elif not quiet:
                die(f"ERROR: {e}")
            else:
                sys.exit(1)
        except ValueError:
            if not quiet:
                err("ERROR: invalid response")
                continue
        except KeyboardInterrupt:
            # Handle Ctrl+C during DNS query
            shared.shutdown = True
            break
        else:
            # Use perf_counter() measurements for accurate wall-clock time
            elapsed = (etime - stime) * 1000  # Convert seconds to milliseconds
            response_time.append(elapsed)
            if not quiet:
                extras = ""
                extras += " %s" % dns.rcode.to_text(answers.rcode())  # add response code

                if show_ttl:
                    if answers.answer:
                        ans_ttl = str(answers.answer[0].ttl)
                        extras += " [TTL=%-4s]" % ans_ttl

                if show_flags:
                    ans_flags = dns.flags.to_text(answers.flags)
                    edns_flags = dns.flags.edns_to_text(answers.ednsflags)
                    if want_dnssec and not (answers.flags & dns.flags.AD):
                        ans_flags += " --"  # add padding to printer output when dnssec is requested, but AD flag is not set
                    extras += " [%s]" % " ".join([ans_flags, edns_flags]).rstrip(' ')  # show both regular + edns flags

                # Display EDNS options compactly (only for explicitly enabled options)
                edns_parts = []

                if want_nsid and answers.options:
                    for ans_opt in answers.options:
                        if ans_opt.otype == dns.edns.OptionType.NSID:
                            nsid_val = ans_opt.nsid
                            if nsid_val:
                                edns_parts.append("NSID:%s" % nsid_val.decode("utf-8"))

                # Always show EDE when present (server-initiated error responses)
                if answers.options:
                    for ans_opt in answers.options:
                        if ans_opt.otype == dns.edns.EDE:
                            if ans_opt.text:
                                # Truncate EDE text for ping output to prevent display issues
                                truncated_text = ans_opt.text[:50] + "..." if len(ans_opt.text) > 50 else ans_opt.text
                                edns_parts.append("EDE:%d(\"%s\")" % (ans_opt.code, truncated_text))
                            else:
                                edns_parts.append("EDE:%d" % ans_opt.code)

                # Always show ECS if present (since it's typically echoed back when requested)
                if answers.options:
                    for ans_opt in answers.options:
                        if ans_opt.otype == dns.edns.OptionType.ECS:
                            if ans_opt.address:
                                edns_parts.append("ECS:%s/%d/%d" % (ans_opt.address, ans_opt.srclen, ans_opt.scopelen))
                            else:
                                edns_parts.append("ECS:auto")

                # Always show cookies when present (echoed back from server)
                if show_cookie and answers.options:
                    for ans_opt in answers.options:
                        if ans_opt.otype == 10:  # COOKIE
                            cookie_hex = ans_opt.client.hex() + ans_opt.server.hex()
                            # Truncate cookie display in normal mode (max 8 chars)
                            if len(cookie_hex) > 8:
                                cookie_display = cookie_hex[:8] + "..."
                            else:
                                cookie_display = cookie_hex
                            edns_parts.append("COOKIE:%s" % cookie_display)

                if edns_parts:
                    extras += " [%s]" % ", ".join(edns_parts)

                if show_answer and answers.answer:
                    ans = answers.answer[0]
                    rtype = dns.rdatatype.to_text(ans.rdtype)
                    extras += " [RDATA: %s %s]" % (rtype, ans[0])

                print("%-3d bytes from %s: seq=%-3d time=%-7.3f ms %s" % (
                    len(answers.to_wire()), server_display, i, elapsed, extras), flush=True)

            if verbose:
                print(answers.to_text(), flush=True)

                # Display EDNS options if present
                if answers.options:
                    print(";EDNS OPTIONS")
                    for ans_opt in answers.options:
                        option_name = "UNKNOWN"
                        option_details = ""

                        if ans_opt.otype == dns.edns.OptionType.NSID:
                            option_name = "NSID"
                            nsid_val = ans_opt.nsid
                            option_details = nsid_val.decode("utf-8") if nsid_val else ""
                        elif ans_opt.otype == dns.edns.OptionType.ECS:
                            option_name = "ECS"
                            option_details = "family=%d, source=%d, scope=%d, address=%s" % (
                                ans_opt.family, ans_opt.srclen, ans_opt.scopelen,
                                ans_opt.address if ans_opt.address else "None")
                        elif ans_opt.otype == dns.edns.EDE:
                            option_name = "EDE"
                            option_details = "code=%d, text=\"%s\"" % (ans_opt.code, ans_opt.text or "")
                        elif ans_opt.otype == 10:  # COOKIE
                            option_name = "COOKIE"
                            client_hex = ans_opt.client.hex()
                            server_hex = ans_opt.server.hex()
                            option_details = "client=%s, server=%s" % (client_hex, server_hex)
                        elif ans_opt.otype == 11:  # TCP-KEEPALIVE
                            option_name = "TCP-KEEPALIVE"
                            if len(ans_opt.data) >= 2:
                                keepalive_timeout = struct.unpack('!H', ans_opt.data[:2])[0]
                                option_details = "timeout=%ds" % keepalive_timeout
                        elif ans_opt.otype == 12:  # PADDING
                            option_name = "PADDING"
                            option_details = "length=%d" % len(ans_opt.data)
                        elif ans_opt.otype == 13:  # CHAIN
                            option_name = "CHAIN"
                            option_details = "closest_encloser=%s" % ans_opt.data.decode('utf-8', errors='ignore')
                        elif ans_opt.otype == 14:  # KEY-TAG
                            option_name = "KEY-TAG"
                            if len(ans_opt.data) >= 2:
                                key_tags = []
                                for idx in range(0, len(ans_opt.data), 2):
                                    if idx + 1 < len(ans_opt.data):
                                        tag = struct.unpack('!H', ans_opt.data[idx:idx + 2])[0]
                                        key_tags.append(str(tag))
                                option_details = "tags=[%s]" % ",".join(key_tags)
                        else:
                            option_details = "type=%d, length=%d" % (ans_opt.otype, len(ans_opt.data))

                        print("%s (%d): %s" % (option_name, ans_opt.otype, option_details), flush=True)

            # Check for shutdown signal after verbose output processing
            if shared.shutdown:
                break

            time_to_next = (stime + interval) - etime
            if time_to_next > 0:
                # Interruptible sleep - check for shutdown signal every 0.1 seconds
                sleep_start = time.time()
                while time.time() - sleep_start < time_to_next:
                    if shared.shutdown:
                        break
                    sleep_duration = time_to_next - (time.time() - sleep_start)
                    if sleep_duration > 0:
                        time.sleep(min(0.1, sleep_duration))

    if tcp_sock:
        try:
            tcp_sock.close()
        except OSError:
            pass

    r_sent = i
    r_received = len(response_time)
    r_lost = r_sent - r_received
    r_lost_percent = (100 * r_lost) / r_sent if r_sent > 0 else 0.0
    if response_time:
        r_min = min(response_time)
        r_max = max(response_time)
        r_avg = sum(response_time) / r_received
        if len(response_time) > 1:
            r_stddev = stdev(response_time)
        else:
            r_stddev = 0
    else:
        r_min = 0
        r_max = 0
        r_avg = 0
        r_stddev = 0

    print('\n--- %s dnsping statistics ---' % server_display, flush=True)
    print('%d requests transmitted, %d responses received, %.0f%% lost' % (r_sent, r_received, r_lost_percent),
          flush=True)
    print('min=%.3f ms, avg=%.3f ms, max=%.3f ms, stddev=%.3f ms' % (r_min, r_avg, r_max, r_stddev), flush=True)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
#
# Copyright (c) 2016-2026, Babak Farrokhi
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pickle
import time
from typing import Any

import cymruwhois

WHOIS_CACHE_FILE = 'whois.cache'


def asn_lookup(ip: str, whois_cache: dict[str, Any]) -> tuple[Any | None, dict[str, Any]]:
    asn: Any | None = None
    try:
        currenttime = time.time()
        if ip in whois_cache:
            asn, ts = whois_cache[ip]
        else:
            ts = 0
        if (currenttime - ts) > 36000:
            c = cymruwhois.Client()
            asn = c.lookup(ip)
            whois_cache[ip] = (asn, currenttime)
    except Exception:
        pass
    return asn, whois_cache


def restore() -> dict[str, Any]:
    try:
        with open(WHOIS_CACHE_FILE, 'rb') as pkl_file:
            whois: dict[str, Any] = pickle.load(pkl_file)
            return whois
    except (IOError, EOFError, pickle.UnpicklingError):
        return {}


def save(whois_data: dict[str, Any]) -> None:
    with open(WHOIS_CACHE_FILE, 'wb') as pkl_file:
        pickle.dump(whois_data, pkl_file)



#!/usr/bin/env python3
#
# Copyright (c) 2016-2026, Babak Farrokhi
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import errno
import socket
import time
from statistics import stdev
from typing import Any

import httpx
import dns.edns
import dns.message
import dns.query
import dns.rcode
import dns.rdataclass

from dnsdiag.shared import random_string, err, unsupported_feature

# Transport protocols
PROTO_UDP: int = 0
PROTO_TCP: int = 1
PROTO_TLS: int = 2
PROTO_HTTPS: int = 3
PROTO_QUIC: int = 4
PROTO_HTTP3: int = 5

_TTL: int | None = None

_PROTO_NAME: dict[int, str] = {
    PROTO_UDP: 'UDP',
    PROTO_TCP: 'TCP',
    PROTO_TLS: 'TLS',
    PROTO_HTTPS: 'HTTPS',
    PROTO_QUIC: 'QUIC',
    PROTO_HTTP3: 'HTTP3',
}

_PROTO_PORT: dict[int, int] = {
    PROTO_UDP: 53,
    PROTO_TCP: 53,
    PROTO_TLS: 853,   # RFC 7858, Section 3.1
    PROTO_HTTPS: 443,
    PROTO_QUIC: 853,  # RFC 9250, Section 4.1.1
    PROTO_HTTP3: 443,
}


class PingResponse:
    def __init__(self) -> None:
        self.r_avg: float = 0.0
        self.r_min: float = 0.0
        self.r_max: float = 0.0
        self.r_stddev: float = 0.0
        self.r_lost_percent: float = 0.0
        self.flags: int = 0
        self.ednsflags: int = 0
        self.ttl: int | None = None
        self.answer: Any | None = None
        self.rcode: int = 0
        self.rcode_text: str = ''
        self.response: Any | None = None


def proto_to_text(proto: int) -> str:
    return _PROTO_NAME[proto]


def get_default_port(proto: int) -> int:
    return _PROTO_PORT[proto]


class CustomSocket(socket.socket):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        # Allow reuse of ports in TIME_WAIT so fixed source ports work across reconnects
        self.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        if _TTL is not None:
            # Set TTL/hop limit based on address family
            if self.family == socket.AF_INET:
                self.setsockopt(socket.SOL_IP, socket.IP_TTL, _TTL)
            elif self.family == socket.AF_INET6:
                self.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_UNICAST_HOPS, _TTL)


def ping(qname: str, server: str, dst_port: int, rdtype: str, timeout: float, count: int, proto: int,
         src_ip: str | None, use_edns: bool = False, force_miss: bool = False,
         want_dnssec: bool = False, want_nsid: bool = False, socket_ttl: int | None = None) -> PingResponse:
    retval = PingResponse()
    retval.rcode_text = "No Response"

    response_times: list[float] = []

    if socket_ttl is not None:
        global _TTL
        _TTL = socket_ttl
        dns.query.socket_factory = CustomSocket

        # Also set QUIC socket factory if available
        # dns.quic._sync.socket_factory is documented as "Can be overridden if needed in special situations"
        try:
            import dns.quic._sync as quic_sync
            quic_sync.socket_factory = CustomSocket
        except (ImportError, AttributeError):
            # aioquic not installed or QUIC not available
            pass

    for i in range(count):

        if force_miss:
            fqdn = "_dnsdiag_%s_.%s" % (random_string(), qname)
        else:
            fqdn = qname

        if use_edns:
            edns_options: list[dns.edns.Option] = []
            if want_nsid:
                edns_options.append(dns.edns.GenericOption(dns.edns.NSID, b''))
            query = dns.message.make_query(fqdn, rdtype, dns.rdataclass.IN, use_edns, want_dnssec, payload=1232,
                                           options=edns_options if edns_options else None)
        else:
            query = dns.message.make_query(fqdn, rdtype, dns.rdataclass.IN, use_edns=False, want_dnssec=False)

        try:
            stime = time.perf_counter()
            if proto == PROTO_UDP:
                response = dns.query.udp(query, server, timeout=timeout, port=dst_port, source=src_ip,
                                         ignore_unexpected=True)
            elif proto == PROTO_TCP:
                response = dns.query.tcp(query, server, timeout=timeout, port=dst_port, source=src_ip)
            elif proto == PROTO_TLS:
                if hasattr(dns.query, 'tls'):
                    response = dns.query.tls(query, server, timeout, dst_port, src_ip)
                else:
                    unsupported_feature()
            elif proto == PROTO_HTTPS:
                if hasattr(dns.query, 'https'):
                    response = dns.query.https(query, server, timeout, dst_port, src_ip,
                                              http_version=dns.query.HTTPVersion.HTTP_2)
                else:
                    unsupported_feature()
            elif proto == PROTO_QUIC:
                if hasattr(dns.query, 'quic'):
                    response = dns.query.quic(query, server, timeout, dst_port, src_ip)
                else:
                    unsupported_feature()
            elif proto == PROTO_HTTP3:
                if hasattr(dns.query, '_http3'):
                    url = f"https://{server}:{dst_port}/dns-query"
                    response = dns.query._http3(query, server, url, timeout, dst_port, src_ip)
                else:
                    unsupported_feature()

        except dns.query.NoDOH:
            raise
        except (httpx.ConnectTimeout, httpx.ReadTimeout,
                httpx.ConnectError):
            raise ConnectionError('Connection failed')
        except ValueError:
            retval.rcode_text = "Invalid Response"
            break
        except dns.exception.Timeout:
            break
        except OSError as e:
            # Transient network errors should be re-raised for caller to handle
            # Exception: during traceroute (socket_ttl set), these errors are expected
            if e.errno in (errno.EHOSTUNREACH, errno.ENETUNREACH):
                if socket_ttl is not None:
                    break
                raise
            elif socket_ttl is not None:  # other acceptable errors while doing traceroute
                break
            err(f"ERROR: {e.strerror}")
            raise OSError(e)
        except Exception as e:
            # In traceroute mode, connection failures are expected
            if socket_ttl is not None:
                break
            err(f"ERROR: {type(e).__name__}: {e}")
            break
        else:
            etime = time.perf_counter()
            # Use perf_counter() measurements for accurate wall-clock time
            elapsed = (etime - stime) * 1000  # Convert seconds to milliseconds
            response_times.append(elapsed)
            if response:
                retval.response = response
                retval.flags = response.flags
                retval.ednsflags = response.ednsflags
                retval.answer = response.answer
                retval.rcode = response.rcode()
                retval.rcode_text = dns.rcode.to_text(response.rcode())
                if len(response.answer) > 0:
                    retval.ttl = response.answer[0].ttl

    r_sent = i + 1
    r_received = len(response_times)
    retval.r_lost_percent = (100 * (r_sent - r_received)) / r_sent
    if response_times:
        retval.r_min = min(response_times)
        retval.r_max = max(response_times)
        retval.r_avg = sum(response_times) / r_received
        if len(response_times) > 1:
            retval.r_stddev = stdev(response_times)
        else:
            retval.r_stddev = 0.0
    else:
        retval.r_min = 0.0
        retval.r_max = 0.0
        retval.r_avg = 0.0
        retval.r_stddev = 0.0

    return retval


def valid_rdatatype(rtype: str) -> bool:
    # validate RR type
    try:
        _ = dns.rdatatype.from_text(rtype)
    except dns.rdatatype.UnknownRdatatype:
        return False
    return True


# Standard DNS flags
_DNS_FLAG_QR = 0x8000
_DNS_FLAG_AA = 0x0400
_DNS_FLAG_TC = 0x0200
_DNS_FLAG_RD = 0x0100
_DNS_FLAG_RA = 0x0080
_DNS_FLAG_AD = 0x0020
_DNS_FLAG_CD = 0x0010

_DNS_FLAGS_BY_TEXT = {
    'QR': _DNS_FLAG_QR,
    'AA': _DNS_FLAG_AA,
    'TC': _DNS_FLAG_TC,
    'RD': _DNS_FLAG_RD,
    'RA': _DNS_FLAG_RA,
    'AD': _DNS_FLAG_AD,
    'CD': _DNS_FLAG_CD
}

# Precompute reverse map and sorted order once
_DNS_FLAGS_BY_VALUE: dict[int, str] = {v: k for k, v in _DNS_FLAGS_BY_TEXT.items()}
_DNS_FLAGS_ORDER: list[tuple[int, str]] = sorted(_DNS_FLAGS_BY_VALUE.items(), reverse=True)


def flags_to_text(flags: int) -> str:
    text_flags = []
    for k, v in _DNS_FLAGS_ORDER:
        if (flags & k) != 0:
            text_flags.append(v)
        else:
            text_flags.append('--')
    return ' '.join(text_flags)



#!/usr/bin/env python3
#
# Copyright (c) 2016-2026, Babak Farrokhi
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import ipaddress
import random
import signal
import socket
import string
import sys
from typing import Any, NoReturn

shutdown: bool = False

__version__ = '2.9.4'


def signal_handler(sig: int, frame: Any) -> None:
    global shutdown
    if shutdown:  # pressed twice, so exit immediately
        sys.exit(0)
    shutdown = True  # pressed once, exit gracefully


def setup_signal_handler() -> None:
    global shutdown
    shutdown = False
    try:
        if hasattr(signal, 'SIGTSTP'):
            signal.signal(signal.SIGTSTP, signal.SIG_IGN)  # ignore CTRL+Z
        signal.signal(signal.SIGINT, signal_handler)  # custom CTRL+C handler
    except AttributeError:  # not all signals are supported on all platforms
        pass


def random_string(min_length: int = 5, max_length: int = 10) -> str:
    char_set = string.ascii_letters + string.digits
    length = random.randint(min_length, max_length)
    return ''.join(random.choices(char_set, k=length))


def die(s: str, exit_code: int = 1) -> NoReturn:
    err(s)
    sys.exit(exit_code)


def err(s: str) -> None:
    print(s, file=sys.stderr, flush=True)


def parse_ip_address(value: str, family: int | None = None) -> str:
    """Parse and validate an IP address string, optionally checking address family.

    family should be socket.AF_INET or socket.AF_INET6 when the caller needs to
    enforce a specific version; leave as None to accept either.
    """
    try:
        addr = ipaddress.ip_address(value)
    except ValueError:
        die(f"ERROR: invalid IP address: {value}")
    if family == socket.AF_INET and not isinstance(addr, ipaddress.IPv4Address):
        die(f"ERROR: expected an IPv4 address, got: {value}")
    elif family == socket.AF_INET6 and not isinstance(addr, ipaddress.IPv6Address):
        die(f"ERROR: expected an IPv6 address, got: {value}")
    return value


def resolve_server_address(dnsserver: str, af: int | None) -> str:
    """Resolve a DNS server hostname to an IP address, or return the IP unchanged.

    For IPv6 resolution, uses AI_V4MAPPED so that IPv4 addresses are also returned.
    Dies with an error message on any resolution failure.
    """
    original = dnsserver
    try:
        ipaddress.ip_address(dnsserver)
        return dnsserver  # already an IP
    except ValueError:
        pass  # hostname — fall through to resolution

    try:
        if af == socket.AF_INET6:
            results = socket.getaddrinfo(dnsserver, None, af, socket.SOCK_DGRAM,
                                         flags=socket.AI_V4MAPPED)
        else:
            results = socket.getaddrinfo(dnsserver, None, af or socket.AF_UNSPEC, socket.SOCK_DGRAM)

        if not results:
            die(f'ERROR: cannot resolve hostname: {original}')

        # getaddrinfo returns (family, type, proto, canonname, sockaddr)
        # sockaddr is (address, port) for IPv4, (address, port, flow, scope) for IPv6
        sockaddr = results[0][4]
        if sockaddr and len(sockaddr) >= 1:
            return str(sockaddr[0])
        die(f'ERROR: invalid address data for hostname: {original}')

    except (OSError, socket.gaierror) as e:
        die(f'ERROR: cannot resolve hostname {original}: {e}')
    except (IndexError, TypeError, ValueError) as e:
        die(f'ERROR: invalid address format for hostname {original}: {e}')


def set_protocol_exclusive(new_proto: int, current_option: str, proto_option_set: str | None) -> tuple[int, str]:
    """Enforce mutual exclusivity of protocol flags, dying if one was already set."""
    if proto_option_set is not None:
        die(f"ERROR: cannot use {current_option} with {proto_option_set}")
    return new_proto, current_option


def unsupported_feature(feature: str = "") -> None:
    if feature:
        die(f"{feature} not available", exit_code=127)
    else:
        die("feature not available", exit_code=127)


def valid_hostname(hostname: str, allow_underscore: bool = False) -> bool:
    """
    Validate a hostname or FQDN.

    For strict RFC 1123/952 validation (hostnames), underscores are rejected.
    For DNS queries (allow_underscore=True), underscores are allowed to support
    special DNS records like _dmarc, _acme-challenge, etc.

    Args:
        hostname: The hostname or FQDN to validate
        allow_underscore: If True, allow underscores in labels (for DNS queries)

    Returns:
        bool: True if hostname is valid, False otherwise
    """
    if not hostname or not isinstance(hostname, str):
        return False

    # Remove trailing dot if present (FQDN notation)
    if hostname.endswith('.'):
        hostname = hostname[:-1]

    # Check overall length (max 253 characters for FQDN)
    if len(hostname) > 253:
        return False

    # Empty after removing trailing dot
    if not hostname:
        return False

    # Split into labels
    labels = hostname.split('.')

    # Each label must be 1-63 characters
    for label in labels:
        if not label or len(label) > 63:
            return False

        # Label must start with alphanumeric or underscore (if allowed)
        if allow_underscore:
            if not (label[0].isalnum() or label[0] == '_'):
                return False
        else:
            if not label[0].isalnum():
                return False

        # Label must end with alphanumeric
        if not label[-1].isalnum():
            return False

        # Label can contain alphanumeric, hyphens, and optionally underscores
        allowed_chars = {'-', '_'} if allow_underscore else {'-'}
        for char in label:
            if not (char.isalnum() or char in allowed_chars):
                return False

    return True


class Colors:
    N = '\033[m'  # native
    R = '\033[31m'  # red
    G = '\033[32m'  # green
    O = '\033[33m'  # orange
    B = '\033[34m'  # blue

    def __init__(self, mode: bool) -> None:
        if not mode:
            self.N = ''
            self.R = ''
            self.G = ''
            self.O = ''
            self.B = ''



#!/usr/bin/env python3
#
# Copyright (c) 2016-2026, Babak Farrokhi
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import concurrent.futures
import getopt
import ipaddress
import os
import socket
import sys
import time
from typing import Any

import dns.edns
import dns.query
import dns.rdatatype
import dns.resolver

import dnsdiag.whois
from dnsdiag.dns import PROTO_UDP, PROTO_TCP, PROTO_QUIC, PROTO_HTTP3, get_default_port
import dnsdiag.shared as shared
from dnsdiag.shared import __version__, Colors, valid_hostname, die, err, set_protocol_exclusive, parse_ip_address, \
    setup_signal_handler, resolve_server_address

# Global Variables
quiet = False
whois_cache: dict[str, Any] = {}

# Constants
__author__ = 'Babak Farrokhi (babak@farrokhi.net)'
__license__ = 'BSD'
__progname__ = os.path.basename(sys.argv[0])


def test_import() -> None:
    #  passing this test means imports were successful
    pass


def usage(exit_code: int = 0) -> None:
    print("""%s version %s
Usage: %s [-aenqhCxTSQ346] [-s server] [-p port] [-c count] [-t type] [-w wait] hostname

Options:
  -h, --help        Show this help message
  -q, --quiet       Enable quiet mode: suppress additional information, showing only traceroute output
  -T, --tcp         Use TCP as the transport protocol
  -Q, --quic        Use QUIC as the transport protocol (DoQ)
  -3, --http3       Use HTTP/3 as the transport protocol (DoH3)
  -4, --ipv4        Use IPv4 (only used when server is a hostname)
  -6, --ipv6        Use IPv6 (only used when server is a hostname)
  -x, --expert      Display expert hints, if available
  -a, --asn         Enable AS# lookups for each encountered hop
  -s, --server      Specify the DNS server to use (default: first system resolver)
  -p, --port        Set the DNS server port number (default: 53)
  -S, --srcip       Set the source IP address for the query (default: address of the default network interface)
  -c, --count       Specify the maximum number of hops (default: 30)
  -w, --wait        Set the maximum wait time for a reply, in seconds (default: 2)
  -t, --type        DNS request record type (default: A)
  -C, --color       Enable colorful output
  -e, --edns        Enable EDNS0 (default: disabled)
  -N, --nsid        Enable NSID to retrieve resolver identification (implies EDNS)
  -n                Disable hostname resolution for IP addresses
""" % (__progname__, __version__, __progname__))
    sys.exit(exit_code)


def expert_report(trace_path: list[str], color_mode: bool) -> None:
    color = Colors(color_mode)
    print("\n%s=== Expert Hints ===%s" % (color.B, color.N))
    if len(trace_path) == 0:
        print(" [*] empty trace - should not happen")
        return

    private_network_radius = 4  # number of hops we assume we are still inside our local network
    prev_hop = None
    if len(trace_path) > 1:
        prev_hop = trace_path[-2]

    if len(trace_path) < 2:
        print(
            " %s[*]%s path too short (possible DNS hijacking, unless it is a local DNS resolver)" % (color.R, color.N))
        return

    if prev_hop == '*' and len(trace_path) > private_network_radius:
        print(" %s[*]%s public DNS server is next to an invisible hop (probably a firewall)" % (color.R, color.N))
        return

    if prev_hop and len(trace_path) > private_network_radius and ipaddress.ip_address(prev_hop).is_private:
        print(" %s[*]%s public DNS server is next to a private IP address (possible hijacking)" % (color.R, color.N))
        return

    if prev_hop and len(trace_path) > private_network_radius and ipaddress.ip_address(prev_hop).is_reserved:
        print(" %s[*]%s public DNS server is next to a reserved IP address (possible hijacking)" % (color.R, color.N))
        return

    # no expert info available
    print(" %s[*]%s No expert hint available for this trace" % (color.G, color.N))


def ping(qname: str, server: str, rdtype: str, proto: int, port: int, ttl: int,
         timeout: int, src_ip: str | None, use_edns: bool, want_nsid: bool) -> tuple[bool, float | None, str | None]:
    reached = False
    resp_time = None
    nsid_value = None
    resp = None

    try:
        resp = dnsdiag.dns.ping(qname, server, port, rdtype, timeout, 1, proto, src_ip, use_edns, force_miss=False,
                                want_dnssec=False, want_nsid=want_nsid, socket_ttl=ttl)

    except SystemExit:
        raise
    except Exception as e:
        die(f"unexpected error: {e}")
    else:
        if resp and resp.answer:
            reached = True
            resp_time = resp.r_max

            # Extract NSID if requested and available
            if want_nsid and resp.response and resp.response.options:
                for option in resp.response.options:
                    if option.otype == dns.edns.OptionType.NSID:
                        nsid_bytes = option.nsid
                        if nsid_bytes:
                            try:
                                nsid_value = nsid_bytes.decode("utf-8")
                            except UnicodeDecodeError:
                                nsid_value = nsid_bytes.hex()
                        break

    return reached, resp_time, nsid_value


def main() -> None:
    global quiet, whois_cache

    setup_signal_handler()

    if len(sys.argv) == 1:
        usage()

    rdatatype = 'A'
    count = 30
    timeout = 2
    dnsserver = None
    proto = PROTO_UDP
    dest_port = get_default_port(proto)
    use_default_dest_port = True
    src_ip = None
    hops = 0
    as_lookup = False
    expert_mode = False
    should_resolve = True
    use_edns = False
    want_nsid = False
    color_mode = False
    af = None  # auto-detect from server address
    af_ipv4_set = False
    af_ipv6_set = False
    proto_option_set: str | None = None

    args = None
    try:
        opts, args = getopt.getopt(sys.argv[1:], "aqhc:s:S:t:w:p:nexCTQ346N",
                                   ["help", "count=", "server=", "quiet", "type=", "wait=", "asn", "port=", "expert",
                                    "color", "srcip=", "tcp", "quic", "http3", "ipv4", "ipv6", "nsid"])
    except getopt.GetoptError as getopt_err:
        err(str(getopt_err))
        usage(1)

    for o, a in opts:
        if o in ("-h", "--help"):
            usage()

    if args and len(args) == 1:
        qname = args[0]
        if not valid_hostname(qname, allow_underscore=True):
            die(f"ERROR: invalid hostname: {qname}")
    else:
        usage(1)

    for o, a in opts:
        if o in ("-c", "--count"):
            try:
                count = int(a)
                if count < 1:
                    die(f"ERROR: count must be positive: {a}")
            except ValueError:
                die(f"ERROR: invalid count value: {a}")
        elif o in ("-x", "--expert"):
            expert_mode = True
        elif o in ("-s", "--server"):
            dnsserver = a
        elif o in ("-q", "--quiet"):
            quiet = True
        elif o in ("-S", "--srcip"):
            src_ip = parse_ip_address(a)
        elif o in ("-w", "--wait"):
            try:
                timeout = int(a)
                if timeout < 0:
                    die(f"ERROR: wait time must be non-negative: {a}")
            except ValueError:
                die(f"ERROR: invalid wait time value: {a}")
        elif o in ("-t", "--type"):
            rdatatype = a
        elif o in ("-p", "--port"):
            try:
                dest_port = int(a)
                if not (0 < dest_port <= 65535):
                    die(f"ERROR: port must be between 1 and 65535: {a}")
                use_default_dest_port = False
            except ValueError:
                die(f"ERROR: invalid port value: {a}")
        elif o in ("-C", "--color"):
            color_mode = True
        elif o == "-n":
            should_resolve = False
        elif o in ("-T", "--tcp"):
            proto, proto_option_set = set_protocol_exclusive(PROTO_TCP, o, proto_option_set)
            if use_default_dest_port:
                dest_port = get_default_port(proto)
        elif o in ("-Q", "--quic"):
            proto, proto_option_set = set_protocol_exclusive(PROTO_QUIC, o, proto_option_set)
            if use_default_dest_port:
                dest_port = get_default_port(proto)
        elif o in ("-3", "--http3"):
            proto, proto_option_set = set_protocol_exclusive(PROTO_HTTP3, o, proto_option_set)
            if use_default_dest_port:
                dest_port = get_default_port(proto)
        elif o in ("-4", "--ipv4"):
            if af_ipv6_set:
                die("ERROR: cannot specify both -4 and -6")
            af = socket.AF_INET
            af_ipv4_set = True
        elif o in ("-6", "--ipv6"):
            if af_ipv4_set:
                die("ERROR: cannot specify both -4 and -6")
            af = socket.AF_INET6
            af_ipv6_set = True
        elif o in ("-a", "--asn"):
            as_lookup = True
        elif o in ("-e", "--edns"):
            use_edns = True
        elif o in ("-N", "--nsid"):
            use_edns = True
            want_nsid = True

    color = Colors(color_mode)

    # validate RR type
    if not dnsdiag.dns.valid_rdatatype(rdatatype):
        die(f'ERROR: invalid record type "{rdatatype}"')

    # Use system DNS server if parameter is not specified
    # remember not all systems have /etc/resolv.conf (i.e. Android)
    if dnsserver is None:
        nameservers = dns.resolver.get_default_resolver().nameservers
        # If user specified -4 or -6, filter for that address family
        if af is not None:
            filtered = []
            for ns in nameservers:
                try:
                    addr = ipaddress.ip_address(str(ns))
                    is_matching = ((af == socket.AF_INET and isinstance(addr, ipaddress.IPv4Address)) or
                                   (af == socket.AF_INET6 and isinstance(addr, ipaddress.IPv6Address)))
                    if is_matching:
                        filtered.append(str(ns))
                except ValueError:
                    pass
            if not filtered:
                af_name = "IPv4" if af == socket.AF_INET else "IPv6"
                die(f"ERROR: no {af_name} nameservers found in system resolver")
            dnsserver = filtered[0]
        else:
            dnsserver = str(nameservers[0])

    # Detect address family from IP, or resolve hostname and default af to IPv4
    try:
        addr = ipaddress.ip_address(dnsserver)
        if isinstance(addr, ipaddress.IPv4Address):
            if af is not None and af != socket.AF_INET:
                die("ERROR: DNS server is IPv4 but -6 flag was specified")
            af = socket.AF_INET
        else:
            if af is not None and af != socket.AF_INET6:
                die("ERROR: DNS server is IPv6 but -4 flag was specified")
            af = socket.AF_INET6
    except ValueError:  # hostname — resolve it
        if af is None:
            af = socket.AF_INET
        dnsserver = resolve_server_address(dnsserver, af)

    # Validate source IP address family against the resolved server family
    if src_ip:
        parse_ip_address(src_ip, family=af)

    # At this point dnsserver must be set (either from command line or system resolver)
    assert dnsserver is not None

    # Select correct ICMP protocol based on address family
    if af == socket.AF_INET:
        icmp_proto = socket.getprotobyname('icmp')
    else:  # AF_INET6
        icmp_proto = socket.getprotobyname('ipv6-icmp')

    ttl = 1
    reached = False
    trace_path = []

    if not quiet:
        # Wrap IPv6 addresses in brackets for better readability
        server_display = f"[{dnsserver}]" if ':' in dnsserver else dnsserver
        print("%s DNS: %s:%d, hostname: %s, rdatatype: %s" % (__progname__, server_display, dest_port, qname, rdatatype),
              flush=True)

    while True:
        # Check for shutdown signal
        if shared.shutdown:
            break

        # some platforms permit opening a DGRAM socket for ICMP without root permission
        # if not availble, we will fall back to RAW which explicitly requires root permission
        try:
            icmp_socket = socket.socket(af, socket.SOCK_RAW, icmp_proto)
        except OSError:
            try:
                icmp_socket = socket.socket(af, socket.SOCK_DGRAM, icmp_proto)
            except OSError:
                die("ERROR: unable to create ICMP socket with unprivileged user. Please run as root.")

        # Bind socket based on address family
        if af == socket.AF_INET:
            icmp_socket.bind(("", dest_port))
        else:  # AF_INET6
            icmp_socket.bind(("::", dest_port))
        icmp_socket.settimeout(timeout)

        curr_addr = None

        with concurrent.futures.ThreadPoolExecutor(max_workers=1) as pool:  # dispatch dns lookup to another thread
            stime = time.perf_counter()
            thr = pool.submit(ping, qname, dnsserver, rdatatype, proto, dest_port, ttl, timeout, src_ip=src_ip,
                              use_edns=use_edns, want_nsid=want_nsid)

            try:  # expect ICMP response
                packet, curr_addr = icmp_socket.recvfrom(512)
                # Parse ICMP packet based on address family
                if af == socket.AF_INET:
                    # IPv4: IP header (20 bytes) + ICMP header (8 bytes) + IP header (20 bytes) + UDP header
                    # ICMP Time Exceeded type = 11
                    if len(packet) > 51:
                        icmp_type = packet[20]
                        l4_dst_port = packet[50] << 8 | packet[51]
                        if icmp_type == 11 and l4_dst_port == dest_port:
                            curr_addr = curr_addr[0]
                        else:
                            curr_addr = None
                    else:
                        curr_addr = None
                else:  # AF_INET6
                    # IPv6: kernel strips IPv6 header, so we get:
                    # ICMPv6 header (8 bytes) + Original IPv6 header (40 bytes) + UDP header
                    # ICMPv6 Time Exceeded type = 3
                    if len(packet) > 50:
                        icmp_type = packet[0]
                        l4_dst_port = packet[50] << 8 | packet[51]
                        if icmp_type == 3 and l4_dst_port == dest_port:
                            curr_addr = curr_addr[0]
                        else:
                            curr_addr = None
                    else:
                        curr_addr = None
            except socket.error:
                pass
            except KeyboardInterrupt:
                shared.shutdown = True
                break
            except SystemExit:
                shared.shutdown = True
                break
            finally:
                etime = time.perf_counter()
                icmp_socket.close()

        try:
            reached, resp_time, nsid_value = thr.result()
        except SystemExit:
            shared.shutdown = True
            break
        except KeyboardInterrupt:
            shared.shutdown = True
            break

        if reached and resp_time is not None:
            curr_addr = dnsserver
            elapsed = resp_time
        else:
            elapsed = abs(etime - stime) * 1000  # convert to milliseconds
            nsid_value = None

        curr_name = curr_addr
        if should_resolve:
            try:
                if curr_addr:
                    curr_name = socket.gethostbyaddr(curr_addr)[0]
            except socket.error:
                curr_name = curr_addr
            except KeyboardInterrupt:
                shared.shutdown = True
                break
            except SystemExit:
                shared.shutdown = True
                break
            except Exception:
                print("unexpected error: ", sys.exc_info()[0])

        # Check for shutdown signal after hostname resolution
        if shared.shutdown:
            break

        if curr_addr:
            as_name = ""
            if as_lookup:
                asn, whois_cache = dnsdiag.whois.asn_lookup(curr_addr, whois_cache)
                as_name = ''
                try:
                    if asn and asn.asn != "NA":
                        as_name = "[AS%s %s] " % (asn.asn, asn.owner)
                except AttributeError:
                    if shared.shutdown:
                        sys.exit(0)

            c = color.N  # default
            if curr_addr != '*':
                try:
                    IP = ipaddress.ip_address(curr_addr)
                    if IP.is_private:
                        c = color.R
                    if IP.is_reserved:
                        c = color.B
                    if curr_addr == dnsserver:
                        c = color.G
                except Exception:
                    pass

            nsid_display = ""
            if nsid_value:
                nsid_display = "[NSID: %s] " % nsid_value

            print("%d\t%s (%s%s%s) %s%s%.3f ms" % (ttl, curr_name, c, curr_addr, color.N, as_name, nsid_display, elapsed), flush=True)
            trace_path.append(curr_addr)
        else:
            print("%d\t *" % ttl, flush=True)
            trace_path.append("*")

        ttl += 1
        hops += 1
        if (hops >= count) or (curr_addr == dnsserver) or reached:
            break

    if expert_mode and not shared.shutdown:
        expert_report(trace_path, color_mode)


if __name__ == '__main__':
    try:
        whois_cache = dnsdiag.whois.restore()
        main()
    finally:
        dnsdiag.whois.save(whois_cache)



#!/usr/bin/env python3
#
# Copyright (c) 2016-2026, Babak Farrokhi
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import concurrent.futures
import datetime
import getopt
import ipaddress
import json
import os
import socket
import sys
import threading
import time

import dns.flags
import dns.rcode
import dns.rdatatype
import dns.resolver

import dnsdiag.dns
from dnsdiag.dns import PROTO_UDP, PROTO_TCP, PROTO_TLS, PROTO_HTTPS, PROTO_QUIC, PROTO_HTTP3, flags_to_text, get_default_port
import dnsdiag.shared as shared
from dnsdiag.shared import __version__, Colors, valid_hostname, die, err, set_protocol_exclusive, parse_ip_address, \
    setup_signal_handler

__author__ = 'Babak Farrokhi (babak@farrokhi.net)'
__license__ = 'BSD'
__progname__ = os.path.basename(sys.argv[0])
print_lock = threading.Lock()


def _resolve_server(server: str) -> str | None:
    """Resolve a server name to an IP, returning None on any failure."""
    try:
        ipaddress.ip_address(server)
        return server
    except ValueError:
        pass
    try:
        results = socket.getaddrinfo(server, None, socket.AF_UNSPEC, socket.SOCK_DGRAM)
        if results and len(results[0]) > 4 and results[0][4]:
            return str(results[0][4][0])
    except (OSError, IndexError, TypeError):
        pass
    return None


def usage(exit_code: int = 0) -> None:
    print("""%s version %s
Usage: %s [-ehmvCTXHQ3SD] [-f server-list] [-j output.json] [-c count] [-t type] [-p port] [-w wait] hostname

  -h, --help         Display this help message
  -f, --file         Specify a DNS server list file to use (default: system resolvers)
  -c, --count        Number of requests to send (default: 10)
  -m, --cache-miss   Force a cache miss measurement by prepending a random hostname
  -w, --wait         Set the maximum wait time for a reply in seconds (default: 2)
  -t, --type         Set the DNS request record type (default: A)
  -T, --tcp          Use TCP as the transport protocol instead of UDP
  -X, --tls          Use TLS as the transport protocol (DoT)
  -Q, --quic         Use QUIC as the transport protocol (DoQ)
  -H, --doh          Use HTTPS as the transport protocol (DoH)
  -3, --http3        Use HTTP/3 as the transport protocol (DoH3)
  -j, --json         Save the results to a specified file in JSONL format (one JSON object per line)
  -p, --port         Specify the DNS server port number (default: protocol-specific)
  -S, --srcip        Set the query source IP address
  -e, --edns         Enable EDNS0 in requests
  -D, --dnssec       Enable the 'DNSSEC desired' (DO flag) in requests
  -C, --color        Enable colorful output
  -v, --verbose      Print the full DNS response details
      --skip-warmup  Disable cache warmup (default: warmup enabled)
""" % (__progname__, __version__, __progname__))
    sys.exit(exit_code)


def maxlen(names: list[str]) -> int:
    return max(len(name) for name in names) if names else 0


def evaluate_server(server: str, qname: str, rdatatype: str, waittime: int, count: int, proto: int,
                    dst_port: int, src_ip: str | None, use_edns: bool, force_miss: bool, want_dnssec: bool,
                    width: int, color: Colors, verbose: bool, json_output: bool, json_filename: str) -> str:
    if not server.strip():
        return ""

    server = server.replace(' ', '')
    resolver = _resolve_server(server)
    if resolver is None:
        return 'ERROR: cannot resolve hostname: %s' % server

    try:
        retval = dnsdiag.dns.ping(qname, resolver, dst_port, rdatatype, waittime, count, proto, src_ip,
                                  use_edns=use_edns, force_miss=force_miss, want_dnssec=want_dnssec)

    except (KeyboardInterrupt, SystemExit):
        raise
    except Exception as e:
        return '%s: %s' % (server, e)

    resolver = server.ljust(width + 1)
    text_flags = flags_to_text(retval.flags)
    edns_flags_text = dns.flags.edns_to_text(retval.ednsflags)
    if edns_flags_text:
        text_flags = " ".join([text_flags, edns_flags_text])
    else:
        text_flags = " ".join([text_flags, "--"])

    s_ttl = str(retval.ttl) if retval.ttl is not None else "N/A"

    if retval.r_lost_percent > 0:
        l_color = color.O
    else:
        l_color = color.N

    output_lines = []

    if json_output:
        outer_data = {
            'hostname': qname,
            'data': {
                'hostname': qname,
                'timestamp': str(datetime.datetime.now()),
                'resolver': resolver.rstrip(),
                'r_min': retval.r_min,
                'r_avg': retval.r_avg,
                'r_max': retval.r_max,
                'r_stddev': retval.r_stddev,
                'r_lost_percent': retval.r_lost_percent,
                's_ttl': s_ttl,
                'text_flags': text_flags,
                'flags': retval.flags,
                'ednsflags': retval.ednsflags,
                'rcode': retval.rcode,
                'rcode_text': retval.rcode_text,
            }
        }

        if json_filename == '-':
            output_lines.append(json.dumps(outer_data))
        else:
            with print_lock:
                with open(json_filename, 'a+') as outfile:
                    json.dump(outer_data, outfile)
                    outfile.write('\n')

    else:
        result = "%s  %-7.2f  %-7.2f  %-7.2f  %-10.2f  %s%%%-3d%s     %-7s  %-26s  %-12s" % (
            resolver, retval.r_avg, retval.r_min, retval.r_max, retval.r_stddev, l_color, retval.r_lost_percent,
            color.N, s_ttl, text_flags, retval.rcode_text)
        output_lines.append(result.rstrip())

    if verbose and retval.answer and not json_output:
        ans_index = 1
        for answer in retval.answer:
            output_lines.append("Answer %d [ %s%s%s ]" % (ans_index, color.G, answer, color.N))
            ans_index += 1
        output_lines.append("")

    return "\n".join(output_lines) if output_lines else ""


def main() -> None:
    setup_signal_handler()

    if len(sys.argv) == 1:
        usage()

    # defaults
    rdatatype = 'A'
    proto = PROTO_UDP
    src_ip = None
    dst_port = get_default_port(proto)
    use_default_dst_port = True
    count = 10
    waittime = 2
    inputfilename: str = ""
    fromfile = False
    json_output = False
    use_edns = False
    want_dnssec = False
    force_miss = False
    verbose = False
    color_mode = False
    warmup = True
    proto_option_set: str | None = None
    qname = 'wikipedia.org'

    try:
        opts, args = getopt.getopt(sys.argv[1:], "hf:c:t:w:S:TevCmXHQ3Dj:p:",
                                   ["help", "file=", "count=", "type=", "wait=", "json=", "tcp", "edns", "verbose",
                                    "color", "cache-miss", "srcip=", "tls", "doh", "quic", "http3", "dnssec", "port=",
                                    "skip-warmup"])
    except getopt.GetoptError as getopt_err:
        err(str(getopt_err))
        usage(1)

    for o, a in opts:
        if o in ("-h", "--help"):
            usage()

    if args and len(args) == 1:
        qname = args[0]
        if not valid_hostname(qname, allow_underscore=True):
            die(f"ERROR: invalid hostname: {qname}")
    else:
        usage(1)

    for o, a in opts:
        if o in ("-c", "--count"):
            try:
                count = int(a)
                if count < 1:
                    die(f"ERROR: count must be positive: {a}")
            except ValueError:
                die(f"ERROR: invalid count value: {a}")
        elif o in ("-f", "--file"):
            inputfilename = a
            fromfile = True
        elif o in ("-w", "--wait"):
            try:
                waittime = int(a)
                if waittime < 0:
                    die(f"ERROR: wait time must be non-negative: {a}")
            except ValueError:
                die(f"ERROR: invalid wait time value: {a}")
        elif o in ("-m", "--cache-miss"):
            force_miss = True
        elif o in ("-t", "--type"):
            rdatatype = a
        elif o in ("-T", "--tcp"):
            proto, proto_option_set = set_protocol_exclusive(PROTO_TCP, o, proto_option_set)
            if use_default_dst_port:
                dst_port = get_default_port(proto)
        elif o in ("-S", "--srcip"):
            src_ip = parse_ip_address(a)
        elif o in ("-j", "--json"):
            json_output = True
            json_filename = a
        elif o in ("-e", "--edns"):
            use_edns = True
        elif o in ("-D", "--dnssec"):
            want_dnssec = True
            use_edns = True
        elif o in ("-C", "--color"):
            color_mode = True
        elif o in ("-v", "--verbose"):
            verbose = True
        elif o in ("-X", "--tls"):
            proto, proto_option_set = set_protocol_exclusive(PROTO_TLS, o, proto_option_set)
            if use_default_dst_port:
                dst_port = get_default_port(proto)
        elif o in ("-H", "--doh"):
            proto, proto_option_set = set_protocol_exclusive(PROTO_HTTPS, o, proto_option_set)
            if use_default_dst_port:
                dst_port = get_default_port(proto)
        elif o in ("-Q", "--quic"):
            proto, proto_option_set = set_protocol_exclusive(PROTO_QUIC, o, proto_option_set)
            if use_default_dst_port:
                dst_port = get_default_port(proto)
        elif o in ("-3", "--http3"):
            proto, proto_option_set = set_protocol_exclusive(PROTO_HTTP3, o, proto_option_set)
            if use_default_dst_port:
                dst_port = get_default_port(proto)
        elif o in ("-p", "--port"):
            try:
                dst_port = int(a)
                if not (0 < dst_port <= 65535):
                    die(f"ERROR: port must be between 1 and 65535: {a}")
                use_default_dst_port = False
            except ValueError:
                die(f"ERROR: invalid port value: {a}")
        elif o in ("--skip-warmup",):
            warmup = False

    # validate RR type
    if not dnsdiag.dns.valid_rdatatype(rdatatype):
        die(f'ERROR: invalid record type "{rdatatype}"')

    color = Colors(color_mode)
    server = ""

    try:
        if fromfile:
            if inputfilename == '-':
                # read from stdin
                with sys.stdin as flist:
                    f = flist.read().splitlines()
            else:
                try:
                    with open(inputfilename, 'rt') as flist:
                        f = flist.read().splitlines()
                except Exception as e:
                    die(str(e))
        else:
            f = dns.resolver.get_default_resolver().nameservers

        if len(f) == 0:
            print("ERROR: No nameserver specified")

        # remove blanks, comments, and empty entries
        f = [name.strip() for name in f if name.strip() and not name.strip().startswith('#')]

        width = maxlen(f)
        blanks = (width - 5) * ' '

        if warmup and not json_output:
            print("Warming up DNS caches...")
            for server in f:
                if shared.shutdown:
                    break
                if not server.strip():
                    continue
                server = server.replace(' ', '')
                resolver = _resolve_server(server)
                if resolver is not None:
                    try:
                        dnsdiag.dns.ping(qname, resolver, dst_port, rdatatype, waittime, 1, proto, src_ip,
                                         use_edns=use_edns, force_miss=force_miss, want_dnssec=want_dnssec)
                    except (KeyboardInterrupt, SystemExit):
                        shared.shutdown = True
                        break
                    except Exception:
                        pass
            if not shared.shutdown:
                time.sleep(1)

        if not json_output:
            print('server' + blanks +
                  '  avg(ms)  min(ms)  max(ms)  stddev(ms)  lost(%)  ttl      flags                      response')
            print((95 + width) * '-')

        max_workers = min(len(f), 10)
        with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
            future_to_server = {}
            for server in f:
                if shared.shutdown:
                    break
                future = executor.submit(evaluate_server, server, qname, rdatatype, waittime, count, proto,
                                       dst_port, src_ip, use_edns, force_miss, want_dnssec, width, color,
                                       verbose, json_output, json_filename if json_output else '')
                future_to_server[future] = server

            for future in future_to_server:
                if shared.shutdown:
                    break
                try:
                    result_output = future.result()
                    if result_output:
                        print(result_output, flush=True)
                except (KeyboardInterrupt, SystemExit):
                    shared.shutdown = True
                    break
                except Exception:
                    pass

    except Exception as e:
        die(f'{server}: {e}')


if __name__ == '__main__':
    main()
